import re 
from refyre.fcluster import FileCluster

class VariableParser:
    '''
    A static class parser designed specifically to identify and evaluate expressions
    relating to variables. This class is the most relevant to the
    node.name attribute, and will normally be used to parse it.
    '''
    def extract_variable_data(expression):
        pattern1 = r'^([\w-]+)(?:\[(\d+)?:(\d+)?(?:\:(\d+))?\])?$'
        pattern2 = r'^([\w-]+)\[(\d+)\]$'

        match1 = re.match(pattern1, expression)
        match2 = re.match(pattern2, expression)

        print(match1, match2)

        if match1:
            print('match1')
            name = match1.group(1)
            start = match1.group(2)
            stop = match1.group(3)
            step = match1.group(4)

            return name, start, stop, step

        elif match2:
            print('match2')
            name = match2.group(1)
            start = match2.group(2)
            stop = str(int(start) + 1)
            step = None

            return name, start, stop, step

        else:
            return None

    def get_slice(var, start, stop, step):
        '''
        Input:
            - Var: the FileCluster variable we seek to obtain a slice from
            - Start: the start of the slice, 0 if nothing specified
            - Stop: the stop of the slice, length if nothing specified
            - Step: the step of the slice, 1 if nothing specified 
        
        Returns a slice object involving the three indices
        '''
        assert type(var) == FileCluster.FileCluster

        start, stop, step = 0 if start is None else int(start), len(var) if stop is None else int(stop), 1 if step is None else int(step)
        return slice(start, stop, step), start, stop, step
    
    def __new__(self, expression, variable_dict):
        '''
            Hijack the '__new__' method (usually used in constructors) 
            and make it do our bidding >:)

            Takes in an expression, and returns the name of the variable, and the sequence of variable requested.
        '''

        out_data = VariableParser.extract_variable_data(expression)

        if not out_data:
            raise Exception(f"Expression {out_data} isn't a valid expression.")

        name, start, stop, step = out_data

        if name not in variable_dict:
            raise Exception(f"{name} is not a recognized variable in variable dict {variable_dict}")

        desired_slice, start, stop, step = VariableParser.get_slice(variable_dict[name], start, stop, step)
        variable_slice = variable_dict[name][desired_slice]

        #Return the variable name, and the subset requested
        return name, variable_slice, start, stop, step


import re
from pathlib import Path
from .cogs.lexer_utils import split_string_ignore_quotes, find_missing_separator
from .cogs.LexerToken import Token
import fnmatch

class Lexer:
    '''
        refyre's static lexer 

        Fundamental assumptions:
             - There are 4 spaces / tab. On some systems, there are 8, hence I've made a SPACES_PER_TAB variable that can be adjusted
             - Only one cluster can be specified per line. The lexer is designed to error if someone tries some dumb shit
    '''
    SPACES_PER_TAB = 4

    def preprocess(input_str):
        '''
        Goal: Clean up the new file in preparation 
        for lexical analysis. This means we're going to 
        get rid of the unnecessary clutter.
        '''

        #Remove any multiline comments
        multiline_removal_pattern = r"'''[\s\S]*?'''"
        updated_content = re.sub(multiline_removal_pattern, '', input_str)

        #Remove any single line comments
        singleline_removal_pattern = r'#.*$'
        updated_content = re.sub(singleline_removal_pattern, '', updated_content)

        #Remove any purely whitespace lines
        whitespace_pattern = r'^\s*$'
        updated_content = re.sub(whitespace_pattern, '', updated_content).strip()

        #Remove any stay newlines
        updated_content = re.sub(r'\n+', '\n', updated_content)
        return updated_content

    def extract_line_data(line):
        '''
            Packages all the important line data into a Token()
            object
        '''

        #Extract all data inside the pattern
        #inside_bracket_pattern = r'\[(.*?)\]'
        #inside_bracket_pattern = r'\[(.*?(?=\[(?!.*?\])).*?)\]'
        #inside_bracket_pattern = r'\[(?:[^\[\]]+|(?R))*\]'
        inside_bracket_pattern = r'\[([^[\]]*(?:\[[^[\]]*\])*[^[\]]*)\]'

        #First, ensure the file is properly bracketed.
        p = find_missing_separator(line)
        if p:
            print(f"ERROR: The lexer has found that some of the attributes don't have a '|' between them. This is a necessary, and will directly alter Lexer output if not fixed. Here's the attributes to check for:\n\n ")
            for w in p:
                print(w)

            print(f"\nHere's the line we detected the error on: {line.strip()}")
            print('Make sure that there\'s a bar between all parts. For example, dir="a"name="k" is incorrect, while dir="a"|name="k" is.')
            print('Aborting ...')
            exit(1)


        try:
            matches = re.search(inside_bracket_pattern, line).group(1)
        except AttributeError as e:
            print(f"ERROR: The lexer attempted to parse line {line} but couldn't find a match. Can you make sure that you have opening and closing brackets? ('[',']')")
            print('Aborting ...')
            exit(1)


        match = split_string_ignore_quotes(matches)
        dict_of_args = {}


        #Extract the key and value of each item, while accounting for potential spaces
        extraction_pattern = r'\s*([^=\s]+)\s*=\s*"(.*?)"\s*'
        for arg in match:

            matches = re.match(extraction_pattern, arg)


            try:
                key = matches.group(1)
            except AttributeError as e:
                print("ERROR: it looks like you're missing a key attribute. Check to make sure each attribute and value have an '=' sign between them, and are in \"\".")
                print(f"The error happened on line: {line.strip()}, when we tried to handle match {arg}")
                print('Aborting ...')
                exit(1)

            try:
                val = matches.group(2)
            except AttributeError as e:
                print("ERROR: it looks like you're missing a val attribute. Check to make sure each attribute and value have an '=' sign between them, and are in \"\".")
                print(f"The error happened on line: {line.strip()}, when we tried to handle match {arg}")
                print('Aborting ...')
                exit(1)

            dict_of_args[key] = val

        #Grab the tab counts - we grab spaces, and count total
        space_pattern = r'^(\s*)\[.*\]'

        num_spaces = len(re.match(space_pattern, line).group(1))

        if num_spaces % Lexer.SPACES_PER_TAB != 0:
            raise Exception(f"Uneven number of tabs - we assume {Lexer.SPACES_PER_TAB} spaces / tab. If you have a different number (ex: 8 spaces / tab), modify the SPACES_PER_TAB variable.")

        try:
            t = Token(num_spaces // Lexer.SPACES_PER_TAB, **dict_of_args)
        except TypeError as e:
            print(f"ERROR: The lexer failed to read one of the attributes. It's most likely because you typed one of the attributes wrong. Here's the full error message below.")
            print(e)
            print('Aborting ...')
            exit(1)

        return t

    def lex(input_file):
        '''
            Performs "lexical analysis", a.k.a grabs all the important information 
            from each line
        '''
        if not Path(input_file).exists() or not Path(input_file).is_file():
            raise Exception(f"The input file path to the spec at {input_file} doesn't exist or is a directory!")

        with open(input_file, "r") as f:
            txt = f.read()
            cleaned_lines = Lexer.preprocess(txt)
            cleaned_lines = cleaned_lines.split('\n')

        tokens = []
        for line in cleaned_lines:
            tokens.append(Lexer.extract_line_data(line))
            
        return tokens

    def __new__(self, input_file):
        return Lexer.lex(input_file)



class FileClusterIterator:
    '''
        A simple iterator to run through the variable data
    '''
    def __init__(self, variable):
        self.var = variable
        self.index = -1

    def __next__(self):

        if self.index >= len(self.var):
            raise StopIteration
        
        self.index += 1
        return self.var[self.index]


from refyre.fgraph import FileGraph
from refyre.reader import Lexer, Parser, ExpressionGenerator, VariableParser, PatternGenerator
from refyre.fcluster import FileCluster
from refyre.utils import is_valid_regex, clone_node
from refyre.core import CodeManager, AliasManager

#pathlib
from pathlib import Path

#Garbage collection
import gc

#Regex
import re

#Deleting
import shutil

#Globbing
import fnmatch

class Refyre:

    def __init__(self, input_specs = None, output_specs = None):
        '''
            input_specs: A List[Str] of filepaths to the various specs refyre will target
        '''

        self.variables = {}
        self.file_graph = FileGraph()
        self.code_manager = CodeManager()
        self.alias_manager = AliasManager()

        if input_specs:
            assert isinstance(input_specs, list)
            for spec in input_specs:
                self.add_spec(spec)
        
        if output_specs:
            assert isinstance(output_specs, list)
            for spec in output_specs:
                self.create_spec(spec)

    def __getitem__(self, key):
        '''
            The refyre variables ("FileClusters") can be
            internally referred to through the Refyre object
        '''
        return self.variables[key] if key in self.variables else None
    

    #The Five Fundamental Operations of Refyre
    def __construct(self, input_path, is_output = False, expand_path = ""):
        '''
            Construct Operation:
                - Receives an input spec, and creates an fgraph
        '''


        print('HEEEERE', input_path)
        p = Parser(Lexer(input_path))

        #Monkey patch, until we figure out how we can support directory clustering with output
        return self.__expand(p, path = expand_path)[0] if not is_output else p
    


    def __expand(self, node, path = "", need_to_append = False):
        '''
            Expands all regex directories into a static fgraph.

            This method enables a single cluster to target multiple directories
        '''
        if not node:
            return [None]

        new_path = Path(node.directory) if node.is_root_dir() else Path(path) / node.directory
        assert Path(path).exists(), f"Error, the path {path} doesn't exist"
        print('new path: ', new_path, node.directory, node.pattern)
        
        if (Path(path).is_file()) or (not new_path.exists() and not PatternGenerator.is_valid_regex(node.directory)):
            print('invalid in general')
            return [None]
        
        #If we are in some sort of sub directory, all variable operations must be appends by default
        print("NEED_TO_APPEND", need_to_append)
        if need_to_append and node.name != '' and not node.name.startswith('+'):
            node.name = '+' + node.name

        if not new_path.exists() and PatternGenerator.is_valid_regex(node.directory) and node.directory != '':
            print('valid regex')

            #Create nodes for each of the pattern matches
            ret = []
            print(new_path, new_path.parent)
            for fl in new_path.parent.iterdir():

                print('file', fl)
                print(r'{}'.format(node.directory), fl.as_posix())
                if re.search(r'{}'.format(PatternGenerator(node.directory)), fl.as_posix()) and fl not in ret and fl.is_dir(): #No files will be allowed to be fclusters

                    #Ensure the node name has append mode, so that all the data across gets added
                    if node.name != '' and not node.name.startswith('+'):
                        node.name = '+' + node.name


                    pattern_matched_node = node.copy()
                    assert pattern_matched_node.pattern == node.pattern, f"Copy node has pattern {pattern_matched_node.pattern} while node has pattern {node.pattern}"

                    assert type(fl.name) == str
                    #Update the directory with the name 
                    pattern_matched_node.directory = fl.name 
                    need_to_append = True

                    print('found match', fl.name, need_to_append)

                    ret.append(pattern_matched_node)                 
            
            #Before we go to children, let's handle any imports we must
            print('handling imports', node.imports, node.directory)
            import_fgraph = None
            if node.imports != '' and Path(node.imports).exists():
                print('importing', node.imports)
                import_fgraph = self.__construct(node.imports, expand_path = new_path)
                import_fgraph.is_root = False
                print('imported', import_fgraph)

            #Now, attempt to check for all the node children of the original node
            #We will insert each of the current node's children into the new matched nodes, and recurse on them

            for i, pattern_node in enumerate(ret):
                nchilds = []

                new_path = Path(pattern_node.directory) if pattern_node.is_root_dir() else Path(path) / pattern_node.directory 
                for child in node.children + [import_fgraph]:
                    nchilds.extend(self.__expand(child, new_path, need_to_append = need_to_append))

                pattern_node.children = [c for c in nchilds if c is not None] 

                if node.flags == '*m' and not new_path.exists():
                    print('making uncreated dir')
                    new_path.mkdir(exist_ok = True, parents = True)

                if pattern_node.type == 'git' and pattern_node.link != '':
                    clone_node(pattern_node.link, new_path)

                if pattern_node.alias != '':
                    print('adding', pattern_node.alias)

                    self.alias_manager.add( ExpressionGenerator(pattern_node.alias)(i), Path(new_path.as_posix()), is_pathlib = True)
                    

            
            return ret

        else:
            print('in else')
            #Before we go to children, let's handle any imports we must
            print('handling imports', node.imports, node.directory)
            import_fgraph = None
            if node.imports != '' and Path(node.imports).exists():
                print('importing', node.imports)
                
                import_fgraph = self.__construct(node.imports, is_output = True, need_to_append = need_to_append)
                import_fgraph.is_root = False

                print("EEEEEETKJSDLKJSADL")
            
            nchild = []
            for child in node.children + [import_fgraph]:
                nchild.extend(self.__expand(child, new_path))
            
            node.children = [c for c in nchild  if c is not None]

            if node.flags == '*m' and not new_path.exists():
                new_path.mkdir(exist_ok = True, parents = True)

            if node.type == 'git' and node.link != '':
                clone_node(node.link, new_path)
                
            if node.alias != '':
                print('adding', node.alias)
                print(ExpressionGenerator(node.alias)(1))
                self.alias_manager.add(ExpressionGenerator(node.alias)(1), Path(new_path.as_posix()), is_pathlib = True)

            return [node]

    def __verify(self, node, path = ""):
        '''
            A simple verification method to ensure the basic integrity of a fgraph.

            This method ONLY CHECKS TO SEE IF DIRECTORIES ARE LEGIT. It does nada mas.

            Assumptions:
                - If a node is root, it will be the base directory, and the absolute path will be taken from there.
        '''
        new_path = Path(node.directory) if node.is_root_dir() else Path(path) / node.directory

        print('Testing to see if ', new_path, 'exists')
        if not new_path.exists():
            return False
        
        new_path = new_path.as_posix()
        for child in node.children:
            if not self.__verify(child, new_path):
                return False

        return True
    

    def __activate(self, node , path = "", mode = "normal"):
        '''
            Activates an fgraph. 
            This means that all variables inside will be collected and factored into the refyre instance.

            The mode parameter defines the behaviour of the method.

            mode = "normal" means normal activation. This is done for input specs, a.k.a. for specs from which we want to 
            COLLECT (retrieve data)

            mode = "deploy" kickstarts deploy activation. Do this for any spec in which we want to TRANSFER / OUTPUT variable
            data somewhere.
        '''

        #Updating the path
        new_path = Path(node.directory) if node.is_root_dir() else Path(path) / node.directory

        if not new_path.exists():
            raise Exception(f"Weird issue caught during activation ... an invalid path appeared - {new_path}")

        new_path = new_path.as_posix()

        if node.name != "":
            print('activating', new_path, node.directory, node.name)
            self.__parse_var(node, new_path, mode)

        if node.flags != '':
            #Activate the code manager
            if '*c' in node.flags:
                self.code_manager.add(new_path)


        for child in node.children:
            self.__activate(child, new_path, mode)

    
    def __output(self, node, path = "", mode = "copy"):
        '''
            Outputs an fgraph. 

            By default, this method performs an in-place generation of the file spec. In other words, no directories will be deleted, and it'll try to create 
            any folders that are missing.

            However, if you want the directory structure in the area to look exactly as mentioned in the output spec (no other dirs), you can specify the flag *c 
            using the flags attribute.
        '''
        #Updating the path
        new_path = Path(node.directory) if node.is_root_dir() else Path(path) / node.directory

        print('mode', node.mode)
        self.__create_output(node, new_path, mode if node.mode == '' else node.mode)
        new_path = new_path.as_posix()

        for child in node.children:
            self.__output(child, new_path, "copy")
        

    def __clear(self):
        '''
            Wipes the refyre instance clean.
        '''

        self.variables.clear()
        self.file_graph = FileGraph()
        gc.collect()

    def __parse_var(self, node, path, mode):
        '''
        Parses all the options in the "names" attribute.

        I expect this method to blow up as this codebase grows (it's an options methods, and you never can have enough options!), 
        so once this goes over 100 lines, we'll move
        it into the fgraph directory.
        '''

        if mode == "normal":

            #Currently, this operator pattern supports only + and "" (appending and defining)
            options_pattern = r'^(\+)?(.*)$' 

            match = re.match(options_pattern, node.name)
            operator = match.group(1) or ""
            name = match.group(2)

            #In this case, the user wants to define a variable
            if operator == "" or (operator == "+" and name not in self.variables):
                print(name, 'define', path)
                print('node', node)
                self.variables[name] = FileCluster(input_paths = [path], input_patterns = [node.pattern])

                print('\nactivation init', operator, node.pattern, path, 'var name', name)
                print('result: ', self.variables[name])

            elif operator == "+":
                self.variables[name] += FileCluster(input_paths = [path], input_patterns = [node.pattern])
                print('\nactivation append', operator, node.pattern, path, 'var name', name)
                print('result: ', self.variables[name])

    def __create_output(self, node, path, mode):
        '''
        Parses all the options in the "names" attribute.

        I expect this method to blow up as this codebase grows (it's an options methods, and you never can have enough options!), 
        so once this goes over 100 lines, we'll move
        it into the fgraph directory.

        mode = "copy" - During output generation, the files are copied over from their original directories 
        mode = "cut"  - During output generation, the files are cut over from their original directories

        '''

        #First of all, let's do the work we need to do 
        path.mkdir(parents = True, exist_ok = True)        

        if node.name != "":
            
            #Extract the key information
            print('extract', node.name)
            name, sliced, start, stop, step = VariableParser(node.name, self.variables)

            #The refresher method will eliminate the invalid paths, so let's add back in the new paths at their original positions
            v = self.variables[name].vals()

            if mode == "copy":
                print('copying ', name, 'to', path, sliced)
                sliced = sliced.copy(path)
                print('post copying ', name, 'to', path, sliced)
            
            elif mode == "cut":
                print('cut', name, 'to', path, sliced)
                sliced = sliced.move(path)
                print('post cut', name, 'to', path, sliced)
            

            s_v = sliced.vals()
            print(sliced, v, start, stop, step)
            for ind_s, ind_v in enumerate(range(start, stop, step)):
                print(ind_s, ind_v, s_v, v)
                v[ind_v] = s_v[ind_s]
            
            print('updated vals', v)
            self.variables[name] = FileCluster(values = v, as_pathlib = True)
            print(self.variables[name])

            
    def __post_generate(self, node, path = "", mode = "copy", flags = ""):
        '''
        Performs any post generation cleanup. For example, it handles the *d flag, which deletes anything
        irrelevant. We do this here to give a chance for any variable data to have been used / moved elsewhere.
        '''

        new_path = Path(node.directory) if node.is_root_dir() else Path(path) / node.directory
        
        print('\n\n',new_path, node.flags)

        if node.flags == '*d' or node.flags == "*da" or flags == "*da":

            #We're going to make a list of all the files that should be there, and find all files that aren't in that list 
            print('On', new_path, node.is_root_dir())

            #First, we get a list of all children that aren't roots 
            children = [ (Path(new_path) / pth.directory) for pth in node.children if not pth.is_root_dir()]
            print('non root', children)

            #Then, we grab all the roots
            children += [ Path(pth.directory) for pth in node.children if pth.is_root_dir()]

            print('all', children)

            var_pths = []
            if node.name != '':
                #Extract the key information - handling slicing
                name, sliced = VariableParser(node.name, self.variables)

                #If we need to serialize the node, we do it first
                if node.serialize != '':
                    sliced = sliced.rename(ExpressionGenerator(node.serialize))
                    self.variables[name] += sliced

                #Next we grab all the values from our variables that are a part of the dir
                p = sliced.filter(lambda x : new_path in x.parents)


                print(p)
                var_pths = p.vals()

            print('variable paths', var_pths)

            #Add 'em up, baby
            tot = children + var_pths
            print('Total', tot)

            bad_files = []
            for pth in Path(new_path).iterdir():
                if pth not in tot:
                    bad_files.append(pth)
                
            print('Bad files: ', bad_files)
            
            #Snip ... snip ... *ouch* >_<
            for fl in bad_files:
                print('Deleting', fl)
                if fl.is_dir():
                    shutil.rmtree(fl.as_posix())
                elif fl.is_file():
                    fl.unlink()
                else:
                    print('No clue how to handle this file')


        new_path = new_path.as_posix()

        for child in node.children:
            self.__post_generate(child, new_path, mode, "*da" if node.flags == "*da" else flags)
        

    def add_spec(self, spec_path):
        '''
        Adds a spec to parse
        '''

        #Construct an fgraph instance
        graph_to_add = self.__construct(spec_path, is_output = False)
        print(graph_to_add)
 
        #Verify the fgraph instance, and then add it in
        if self.__verify(graph_to_add):

            print('Verification successful.')
            self.file_graph.add_graph(graph_to_add)

            print('Graph added. Activating variables.\n\n')
            print(self.file_graph)
            print('\n\n')

            self.__activate(self.file_graph.fgraph_root)

            print('On standby. Variables: ')
            print(self.variables)
        else:
            print('Vertification failed. Maybe the spec has an invalid dir parameter specified somewhere?')

        print('Spec addition complete.\n\n')
    
    def create_spec(self, spec_path, mode = "cut", track = False):
        '''
        Creates a spec given by the spec path

        spec_path - filepath to the spec path
        mode - 
            - "cut": Any files that are being transferred to create this spec will be cut from their place
            - "copy": Any files that are being transferred to create this spec will be duplicated from their original place
        '''

        assert Path(spec_path).exists(), f"The spec to create at {spec_path} cannot be located"

        print('Creating spec.')

        #Construct the spec path
        graph_to_add = self.__construct(spec_path, is_output = True)

        print(graph_to_add)
        print('Graph constructed. Activating variables.')

        #Do any actions inside the fgraph
        self.__output(graph_to_add, mode = mode)

        print('Output complete. Commencing cleanup.')

        #Do any cleanup inside the fgraph
        self.__post_generate(graph_to_add, mode = mode, flags = graph_to_add.flags)

        if track:
            print('Tracking requested. Adding to fgraph cluster.')

            self.file_graph.add_graph(graph_to_add)

            print('Graph added.')
            
    def get_vars(self):
        return self.variables



from pathlib import Path
import sys
import venv
import site
from refyre.utils import optional_dependencies
from refyre import Refyre
from pip._internal import main as pip_main
import subprocess

def create_virtualenv_and_install_requirements(venv_path, requirements_file):
    # Create the virtual environment
    venv_builder = venv.EnvBuilder(with_pip=True)
    venv_builder.create(venv_path)

    # Activate the virtual environment
    activate_script = (
        Path(venv_path) / "Scripts" / "activate" if sys.platform == "win32"
        else Path(venv_path) / "bin" / "activate"
    )
    activate_cmd = f"source {activate_script} && python -m pip"
    
    # Install the requirements using pip within the virtual environment
    print("installing reqs")
    pip_install_cmd = f"{activate_cmd} install -r {requirements_file}"
    subprocess.run(pip_install_cmd, shell=True, check=True)

class RecipePreprocessor:
    def __new__(self, recipe_path):
        '''
            Performs all the recipe path preprocessing, and then 
            returns the Refyre instance
            to manage.

            Returns the refyre instance
        '''
        with optional_dependencies("warn"):
            import yaml 

            assert Path(recipe_path).exists(), f"The input file you were trying to access - {recipe_path} - does NOT exist"

            recipe_dict = None
            with open(recipe_path, 'r') as f:
                recipe_dict = yaml.safe_load(f)
            
            '''
            Currently, a recipe consists of attributes:

                input-spec: a filepath to the input spec
                base-dir: filepath to the base directory / entry point you want your program to see
                env: a description of the environment you want to setup, with two attributes
                    - name: name of the environment
                    - requirements: requirements file

            '''

            print(recipe_dict)
            print(recipe_dict['env']['name'])

            #Setup the base directories based on an input spec
            #ref = Refyre.Refyre(input_specs = [recipe_dict["input-spec"]])

            #Create the env if it doesn't exist already:
            if 'env' in recipe_dict and recipe_dict['env']:
                if 'name' in recipe_dict['env']:
                    if not Path(recipe_dict['env']['name']).exists():
                        create_virtualenv_and_install_requirements(recipe_dict['env']['name'], recipe_dict['env']['requirements'])

            
            #Setup the base dir
            return ref
            


import re 





class ExpressionGenerator:
    '''
        The ExpressionGenerator class takes in an 
        input generator string, such as test$.txt, and returns a
        lambda function that can be used to evaluate the expressions at various points.

        For example, test$.txt returns a lambda function, that when called at i = 1, returns test1.txt

        This class is a static class; it's sole purpose is to create an organized location
        to generate complex expressions.

        For now, only numbers are supported. You can generate number expressions using the '$' symbol 
        as shown above.

    '''
    def generate_expression(string):
        # Replace '$' with '{}' to use it as a placeholder for string formatting
        lambda_function = lambda i: string.replace('$', str(i))

        return lambda_function

    def __new__(self, string):
        '''
            Hijack the '__new__' method (usually used in constructors) 
            and make it do our bidding >:)

            Takes in an expression, return a lambda function.
        '''

        return ExpressionGenerator.generate_expression(string)


from refyre.utils import optional_dependencies 
from refyre.fcluster import FileCluster #Time to use the power of our variables :)
from pathlib import Path
import random
import shutil

def clone_node(url, fp):
    import git 

    print("GEET", git)


    #Ensure no file conflicts
    rand_num = ''.join(["{}".format(random.randint(0, 9)) for num in range(0, 20)])
    tmp_empty = fp / f'tmp{rand_num}'


    print('Cloning to ', tmp_empty)
    git.Repo.clone_from(url, tmp_empty)

    mv_var = FileCluster(input_paths = [tmp_empty], input_patterns = ['*'])

    mv_var.move(tmp_empty.parent)

    tmp_empty.rmdir()



import sys 
from pathlib import Path

class CodeManager:

    def __init__(self):
        self.module_dirs = []
        self.base_sys_paths = list(sys.path)

    def add(self, filepath):
        p = Path(filepath)
        if p.is_dir() and filepath not in sys.path:
            print('adding', filepath)

            self.module_dirs.append(filepath) 
            sys.path.append(filepath)
        elif p.is_file() and p.parent.as_posix() in sys.path:
            print('adding parent since it is file', p.parent.as_posix())

            self.module_dirs.append(p.parent.as_posix()) 
            sys.path.append(p.parent.as_posix())



    



from .FileGraphNode import FileGraphNode

class FileGraph:
    '''
        This class is really just a wrapper class.

        The goal is to provide a proxy interface for the Refyre objects to 
        interact with the fgraph without interfering with the inner
        workings.
    '''

    def __init__(self):
        #Make a virtual node stringing together all the graphs 

        #DO NOT DELETE THE children []. This led me on a 2 hour goose chase; i.e. object memory gets repeatedly screwed up, causing it to conflate values with another node. Don't delete it. DONT!!!
        self.fgraph_root = FileGraphNode(is_root = True, children = [])     


    def add_graph(self, graph):
        self.fgraph_root.add_child(graph)
    
    def __repr__(self):
        return repr(self.fgraph_root)
    



from pathlib import Path
from functools import wraps
from refyre.utils import get_optimal_pattern, optional_dependencies
from refyre.reader import PatternGenerator
import re

#Copying / Deleting
import shutil

#Zipping
import zipfile


#Iteration
from .FileClusterIterator import FileClusterIterator
from .AutoRefresher import AutoRefresher

def refresh(input_arr):
    '''
    A simple refresher method. This will handle variable conflicts.

    Whenever data is altered, all the invalid paths will drop out.

    Particularly useful for variables with overlapping data
    '''
    return [ v for v in set(input_arr) if v.exists() ]
        
def with_refresh(func):
    def refreshed(self, *args, **kwargs):
        self.values = refresh(self.values)
        return func(self, *args, **kwargs)

    return refreshed

class FileCluster:
    '''
        The workhorse of refyre. Variables do all of the heavylifting, harnessing
        the power of the data parsed through the fgraphs.

        Paradigms - any method that involves transporting data must be wrapped in the 
        refresher decorator method above. This will prevent overlapping variables from 
        incorrectly altering results from each other, as well as ensures that variables
        only contain valid data.
    '''

    #Counter to manage IDs
    GLOBAL_COUNTER = 0

    #Stores references to each FileCluster; in the future, this can be used for tracking & mem management
    clusters = []

    def __init__(self, input_paths = [], input_patterns = [], values = [], as_pathlib = False):
        assert len(input_paths) == len(input_patterns), "Uneven lengths for input paths and patterns"

        for pattern in input_patterns:
            print('add', pattern, PatternGenerator(pattern))
        self.values = self.__populate(input_paths, [ PatternGenerator(p) for p in input_patterns  ])

        #ID the variable
        self.id = FileCluster.GLOBAL_COUNTER
        FileCluster.GLOBAL_COUNTER += 1

        #Track the cluster
        FileCluster.clusters.append(self)
    
        #If the user didn't send the values as Pathlib objects already
        if not as_pathlib: 
            self.values += [ Path(pth) for pth in values ] 
        else:
            self.values += values

    
    def all_clusters(self):
        return FileCluster.clusters
    
    @AutoRefresher()
    def __add__(self, other):
        '''
            other - another FileCluster

            Returns a file cluster with all the combined values
        '''
        #Ensure the input_data is as desired
        assert type(other) == FileCluster

        new_values = self.values + other.values

        return FileCluster(input_paths = [], input_patterns = [], values = new_values, as_pathlib = True)
    
    @AutoRefresher()
    def __and__(self, other):
        '''
            other - another FileCluster

            Returns the intersection between the two FileClusters
        '''

        return FileCluster(input_paths = [], input_patterns = [], values = list(set(self.values).intersection(other.values)))
    
    @AutoRefresher()
    def __or__(self, other):
        '''
            other - another FileCluster

            Returns the union between two FileClusters
        '''

        return FileCluster(input_paths = [], input_patterns = [], values = list(set(self.values).union(other.values)))
    
    @AutoRefresher()
    def __sub__(self, other):
        '''
            other - another FileCluster

            Returns all the files that are exclusively in self
        '''

        return FileCluster(input_paths = [], input_patterns = [], values = [v for v in self.values if v not in other.values])

    @AutoRefresher()
    def __eq__(self, other):
        return set(self.values) == set(other.values)
    
    @AutoRefresher()
    def contains(self, other):
        return self & other == other

    def __populate(self, input_paths, input_patterns):
        '''
        Populates a variable, using the input paths and directories 
        '''

        ret = []
        for dir_path, pattern in zip(input_paths, input_patterns):
            print('pot', dir_path, pattern)
            for fl in Path(dir_path).iterdir():
                print(pattern, fl.name,re.search(r'{}'.format(pattern), fl.name) )
                if re.search(r'{}'.format(pattern), fl.name) and fl not in ret:
                    print('fl', fl, 'is match to ', dir_path, pattern)
                    ret.append(fl)
        
        return ret

    @AutoRefresher()
    def vals(self):
        '''
            Returns a Pathlib.Paths list of all
            filepaths
        '''
        return self.values
    
    @AutoRefresher()
    def dirs(self):
        '''
            Returns a Paths list of all parent directories
        '''
        return list(set([v.parent for v in self.values]))
    
    @AutoRefresher()
    def __repr__(self):
        return f"FileCluster(values = {self.values})"

    @AutoRefresher()
    def __len__(self):
        return len(self.values)

    @AutoRefresher()
    def __iter__(self):
        return 
    
    @AutoRefresher()
    def __getitem__(self, key):
        print(key)
        if isinstance(key, slice):
            return FileCluster(input_patterns = [] , input_paths = [], values = self.values[key.start:key.stop:key.step])
        elif isinstance(key, int):
            return FileCluster(input_patterns = [] , input_paths = [], values = self.values[key])

    #Update a map_func
    @AutoRefresher()
    def map(self, map_func):
        '''
        Input: 
            - map_func: a function to map each filepath. The mapping must take a filepath and return a filepath
        
        Return:
            - FileCluster object with the mapped values
        '''

        nvals = []
        for v in self.values:
            nvals.append(Path(map_func(v))) #Append a copy of the object to prevent object ref shenanigans
        
        p = FileCluster(input_patterns = [], input_paths = [], values = nvals, as_pathlib = True)
        print(p)
        return p
    
    @AutoRefresher()
    def reduce(self, reducer_function):
        '''
        Input: 
            - reducer_function(a, b): Takes in two parameters that can be reduced into a third parameter
        
        Returns
            - the Result of the reduction, or None if the array is empty
        '''
        if len(self.values):
            a = self.values[0]

            for i in range(1, len(self.values)):
                a = reducer_function(a, self.values[i])

            return a
        
        return None

    def move(self, target_dir):
        '''
        Input: 
            - target_dir: the directory to which all files will be sent.
        
        Returns
            - FileCluster object with the new values 
        '''

        p = Path(target_dir)

        #If the directory doesn't actually exist or just isn't a directory, we return None
        if not p.exists() or not p.is_dir():
            return None
        
        def move_func(v):
            shutil.move(str(v), str(p / v.name))
            return p / v.name

        @AutoRefresher(does_modify = True, mapper_func = lambda x : p / x.name)
        def exec_func(self):
            return self.map(move_func)

        return exec_func(self)
    
    def copy(self, target_dir):
        '''
        Input: 
            - target_dir: the directory to which all files will be sent.
        
        Returns
            - FileCluster object with the new values 
        '''
        p = Path(target_dir)

        #If the directory doesn't actually exist or just isn't a directory, we return None
        if not p.exists() or not p.is_dir():
            return None

        def copy_func(v):
            shutil.copy(str(v), str(p / v.name))
            return p / v.name

        @AutoRefresher(does_modify = True, mapper_func = lambda x : p / x.name, instance = self)
        def exec_func(self):
            return self.map(copy_func) 

        return exec_func(self)
    
    def filter(self, filter_func):
        '''
        Input: 
            - filter_func: a function to filter out values, given an input of a filepath.
                - i.e, return False for any values we don't want to keep
        
        Return:
            - FileCluster object with the filtered values
        '''


        nvals = []
        @AutoRefresher(does_modify = True, does_filter = True, filter_func = filter_func, instance = self)
        def work(self):
            for v in self.values:
                if filter_func(v):
                    nvals.append(Path(v.as_posix())) #Append a copy of the object to prevent object ref shenanigans
        
        work(self)
        return FileCluster(input_patterns = [], input_paths = [], values = nvals, as_pathlib = True)
    
    def delete(self):
        '''
        Deletes all the files in the variable. 

        Return:
            - Empty FileCluster object
        '''

        @AutoRefresher(does_modify = True, does_filter = True, filter_func = lambda x : False)
        def work(self):
            for fl in self.values:
                if fl.is_dir():
                    shutil.rmtree(fl.as_posix())
                elif fl.is_file():
                    fl.unlink()

        work(self)
        return FileCluster(input_patterns = [], input_paths = [], values = [])

    def rename(self, rename_function):
        '''
            rename_function: a function that takes an integer as an input 
        '''

        i = -1
        def sample(x):
            nonlocal i
            i += 1
            return x.parent / rename_function(i)

        @AutoRefresher(does_modify = True, mapper_func = sample)
        def work(self):
            nvals = []
            for i, v in enumerate(self.values):
                print('renaming', v,v.parent / rename_function(i) )
                nvals.append(Path(v.as_posix()).rename(v.parent / rename_function(i)))
            return nvals

        nvals = work(self)
        return FileCluster(input_patterns = [], input_paths = [], values = nvals, as_pathlib = True)

    @AutoRefresher()
    def zip(self, save_dir = '.', save_name = "out.zip"):
        '''
        Zips all the files into a single zip.
        '''

        save_p = Path(save_dir) / save_name

        copied = self.copy(save_dir)

        with zipfile.ZipFile(save_p.as_posix(), 'w') as zipMe:        
            for fl in copied.values:
                zipMe.write(fl.as_posix(), compress_type=zipfile.ZIP_DEFLATED)
        
        assert save_p.exists(), "For some reason, the newly created zip file does not exist"

        copied.delete()
        return FileCluster(input_patterns = [], input_paths = [], values = [save_p], as_pathlib = True)

    @AutoRefresher()
    def post(self, url, additional_data, payload_name):

        with optional_dependencies('warn'):
            #GET/POST requests
            import requests
            '''
            Uses the requests library to send a post request to the url, 
            alongside the additional payload parameters you specify. All the data 
            you want to push through is sent as a zip 

                url: Str link to the API endpoint
                additional_data: a dict to any other metadata you want to ferry across
                payload_name: what you want to name the zip going across


            Return type: the Response object returned from the requests.get() call
            '''

            #Zip all the files up to send
            zipped = self.zip()
            zipped_fp = zipped.vals()[0].as_posix()

            resp = None
            with open(zipped_fp, "rb") as f:
                resp = requests.post(url = url, params = additional_data, files = { payload_name : f })

            #Delete the intermediate zip we constructed
            zipped.delete()

            return resp 

    


import re
import fnmatch

'''
Utility methods for regex related things 
'''

def is_valid_regex(string):
    try:
        re.compile(fnmatch.translate(string))
        return True
    except re.error:
        return False

def get_optimal_pattern(string):
    try:
        print('string is a regex', string )
        re.compile(string)
        return string
    except re.error:
        return False
    
    try:
        print('string is a glob', string )
        re.compile(fnmatch.translate(string))
        return fnmatch.translate(string)
    except re.error:    
        return None



import fnmatch

class Token:
    '''
        A single Lexer token
    '''
    def __init__(self, tab_level, pattern = "", dir = "",  type = "", name = "", flags = "", serialize="", imports="", mode = "", link = "", alias = ""):
        self.tab_level = tab_level
        self.pattern = pattern
        self.directory = dir
        self.dirtype = type
        self.name = name
        self.flags = flags
        self.serialize = serialize
        self.imports = imports
        self.mode = mode
        self.link = link
        self.alias = alias


    def __repr__(self):
        return f"Token(tab_level={self.tab_level},pattern={self.pattern},directory={self.directory},dirtype={self.dirtype},name={self.name},flags={self.flags},serialize={self.serialize})"



from .Lexer import Lexer
from .cogs.LexerToken import Token

from refyre.fgraph import FileGraphNode
from pathlib import Path


#Utility stack (LIFO) data stucture methods
def push(stack, item):
    stack.append(item)

def pop(stack):
    return stack.pop()

class Parser:
    '''
        Refyre's static parser
    '''
    def parse(tokens, INF = -10, ):
        '''
        Things to note: 
            - Floor for tab depth is set to be -10. If it goes beyond this, ...
            the circus would really love to hire their new rising clown.
        '''
        stack = []

        #Push our virtual top
        push(stack, (FileGraphNode(is_root = True, children = []), INF))

        
        for tok in tokens:
            cur = []

            #Resolve all the nodes with tab_level lower
            while tok.tab_level < stack[-1][1]:
                top = pop(stack)

                #Grab all the nodes with the same level
                while stack[-1][1] == top[1]:
                    p = pop(stack)[0]
                    cur = [p] + cur 

                #Append top last to preserve ordering - THIS IS CRUCIAL TO PRESERVE USER INTENT
                cur.append(top[0])
                
                #Now, we know that the next node, whatever it is, must have a lower level
                stack[-1][0].add_children(cur)

                #Clear the list
                cur.clear()

            print('imports', tok.imports, tok.link)
            push(stack, (FileGraphNode(children = list([]), pattern = tok.pattern, directory = tok.directory, type = tok.dirtype, name = tok.name, is_root = Path(tok.directory).exists(), flags = tok.flags, serialize = tok.serialize, imports = tok.imports, mode = tok.mode, link = tok.link, alias = tok.alias) , tok.tab_level))

        #Compress the remainder of the stack
        while len(stack) > 1: 

            top = pop(stack)
            cur = []

            while top[1] == stack[-1][1]:
                #Grab all the nodes with the same level
                p = pop(stack)[0]
                cur = [p] + cur
            
            #Append top last to preserve ordering - THIS IS CRUCIAL TO PRESERVE USER INTENT
            cur.append(top[0])

            #Now, we know that the next node, whatever it is, must have a lower level
            stack[-1][0].add_children(cur)

            #Clear the list
            cur.clear()
        
        assert len(stack) == 1 and stack[0][1] == INF, "Improper root node, something went wrong in the parsing algorithm"
        

        return stack[0][0]

    def __new__(self, tokens):
        assert type(tokens) == list and all([type(t) == Token for t in tokens]), "Invalid type found from Lexer output; something's going wrong."
        return Parser.parse(tokens)


from .Lexer import Lexer
from .Parser import Parser
from .ExpressionGenerator import ExpressionGenerator 
from .PatternGenerator import PatternGenerator
from .cogs.VariableParser import VariableParser 

#Internal 
from .cogs.lexer_utils import split_string_ignore_quotes


import re

def split_string_ignore_quotes(string):
    pattern = r'"[^"]*"'
    matches = re.findall(pattern, string)

    for match in matches:
        string = string.replace(match, match.replace('|', '\x00'))

    splitted = string.split('|')

    for i in range(len(splitted)):
        splitted[i] = splitted[i].replace('\x00', '|').strip()

    return splitted

def find_missing_separator(string):
    pattern = r'"\s*(\w+)=\S[^"]*"([^|\s])'
    matches = re.findall(pattern, string)
    missing_separators = []

    for match in matches:
        missing_separators.append(match[0])

    return missing_separators


from pathlib import Path 
from refyre.reader import PatternGenerator
import copy

class FileGraphNode:
    '''
        How the graph will be orchestrated:

        Problem: There may be multiple root "entry" directories the user many want to be 
        able to target.

        Solution: We abstract that away in the graph.

        We will have one virtual "root" node stringing together all the other entry points.
    '''
    def __init__(self, children = [], pattern = "", directory = "", type = "", name = "", is_root = False, flags = "", serialize = "", imports="", mode = "", link="", alias=""):
        self.children = children

        self.pattern = pattern #fnmatch.translate(pattern) #Add this attribute to handle glob pattern recognition
        self.directory = directory
        self.type = type
        self.name = name
        self.is_root = is_root
        self.flags = flags
        self.serialize = serialize
        self.imports = imports
        self.mode = mode
        self.link = link
        self.alias = alias


    def add_child(self, child):
        assert isinstance(child, FileGraphNode), "Children of FileGraphNode should only be FileGraphNodes"
        
        self.children.append(child)
    
    def add_children(self, children):
        for child in children:
            self.add_child(child)

    def is_root_dir(self):
        return self.is_root
        
    def __repr__(self):
        return f"FileGraphNode(children = {self.children}, pattern = {self.pattern}, directory = {self.directory}, type = {self.type}, name = {self.name}, is_root = {self.is_root}, flag = {self.flags}, serialize = {self.serialize}, imports = {self.imports}, mode = {self.mode})"
    
    def __copy__(self):
        new_obj = FileGraphNode()
        for k,v in vars(self).items():
            try:
                setattr(new_obj, k, copy.deepcopy(v))
            except Exception as e:
                print(e)

        return new_obj
    
    def copy(self):
        return self.__copy__()


from refyre.utils import optional_dependencies 

with optional_dependencies("warn"):
    import numpy as np 
    import pandas as pd
    import matplotlib.pyplot as plt

class PandasStack:

    def __init__(self, variables_array):
        self.variables_array = variables_array
    
    def create_dataframe(self, col_names, mapper_func):
        '''
            col_names: List[String] An array of the names of the columns of the dataframe
            mapper_func: A function that takes in input of a tuple (var1_output, var2_output, ...), and returns a tuple with values for each of the columns
                - Note each tuple value should be the value for the corresponding columns (i.e the ith col_name should be associated with the ith tuple val)
        '''

        vals = [ v.vals() for v in self.variables_array]
        out_dct = {}

        for n in col_names:
            out_dct[n] = []

        for input_tuple in zip(*vals):

            output_data = mapper_func(input_tuple)
            assert len(output_data) == len(col_names), "The data length and number of columns specified through the column names don't match up!"

            for i in range(len(col_names)):
                out_dct[col_names[i]].append(output_data[i])

        return pd.DataFrame.from_dict(out_dct)

class NumpyStack:

    def __init__(self, variables_array):
        self.variables_array = variables_array
    
    def create_numpy(self, processing_function):
        '''
            processing_function: 
        '''
        pass 







from pathlib import Path

class AliasManager:

    def __init__(self):
        self.alias_dict = {}

    def add(self, alias, path, is_pathlib = False):
        self.alias_dict[alias] = Path(path) if not is_pathlib else path
    
    def __getitem__(self, key):
        return self.alias_dict[key]

    def __repr__(self):
        return f"AliasManager(aliases = {self.alias_dict})"
    
    def aliases(self):
        return self.alias_dict


from pathlib import Path  

def insert(fdct, k, v):
    if k not in fdct:
        fdct[k] = set()
    
    fdct[k].add(v)

class AutoRefresher:
    '''
    The AutoRefresher acts as a failsafe, gatekeeping mechanism for 
    the FileCluster.

    Here's what it does - 
        - Ensures that the variables contain unique, valid files. Any duplicates or nonexistent paths are removed.
        - Updates any updated paths for overlapping variables; i.e. if you change a variable's data and move some files, other variables that may share the
        same filepaths will have their filepaths updated

    AutoRefresher currently supports two modes - 
    1. "preserve" - This is the default mode. If any changes are made in one variable's contents, the updates are propogated to the rest of the vars.
    2. "optimize_single_real" - The most optimized, this mode deletes any paths in the variable that no longer exist, and ensures the paths are purely unique

    '''
    REFRESHER_MODE = 'preserve'
    VALID_MODES = ['preserve']

    #A global dictionary to manage filepath movements
    files_dict = {} #Maps filepath strings to new strings. If a filepath is expected to be gone, it will contain the value None

    def __init__(self, *a, **kw):
        self.conf_args = a
        self.conf_kw = kw
        
       
    def __call__(self, func):
        
        if AutoRefresher.REFRESHER_MODE == 'preserve':


            def wrapper(instance, *args, **kwargs):

                #Update & Delete any paths using the files_dict
                nvals = [] 
                fdct = AutoRefresher.files_dict

                #Remove all instances of var.id from dict
                for k in fdct:
                    fdct[k].discard(instance.id)

                clusters = instance.all_clusters()

                for v in instance.values:
                    if v not in nvals and "does_modify" in self.conf_kw and self.conf_kw["does_modify"]:
                        if "mapper_func" in self.conf_kw:

                            '''
                                How can a file conflict occur?

                                No matter what, if the mapped file already exists in the dictionary, then we 
                                say a file conflict exists. 

                                This is quite a strong statement, but renaming a file (false positive) is a lot
                                safer than losing the file entirely (false negative) 
                            '''
                            #Check and handle file conflicts

                            i = 0
                            potential_candidate = v

                            if self.conf_kw["mapper_func"](potential_candidate).as_posix() != potential_candidate.as_posix():
                                while (self.conf_kw["mapper_func"](potential_candidate)).as_posix() in fdct:
                                    print('file conflict', potential_candidate, self.conf_kw["mapper_func"](potential_candidate))
                                    i += 1
                                    potential_candidate = self.conflict_filename(potential_candidate.parent, potential_candidate.stem, potential_candidate.suffix, i) 
                                

                            if potential_candidate != v:
                                #Move all relevant ids to the filename

                                #Remove all the old values
                                for cluster_id in fdct[v.as_posix()]:
                                    cluster_var = clusters[cluster_id]
                                    cluster_var.values = [p for p in cluster_var.values if p.as_posix() != v.as_posix()] + [Path(potential_candidate.as_posix())]

                                o_set = fdct[v.as_posix()]
                                
                                assert potential_candidate.as_posix() not in fdct, "New candidate is somehow in our dictionary"

                                #Create a new set
                                fdct[self.conf_kw["mapper_func"](potential_candidate).as_posix()] = set()

                                #Update dict
                                for n in o_set:
                                    fdct[self.conf_kw["mapper_func"](potential_candidate).as_posix()].add(n)

                                #Remove the old set
                                fdct.pop(v.as_posix(), None)

                                #Rename the file
                                v.rename(potential_candidate)
                                assert potential_candidate.exists(), "New candidate doesnt exist after renaming"

                            #update dict, and add to nvals

                            insert(fdct, self.conf_kw["mapper_func"](potential_candidate).as_posix(), instance.id)
                            nvals.append(potential_candidate)
                        
                        if "does_filter" in self.conf_kw and self.conf_kw["does_filter"] and "filter_func" in self.conf_kw:
                            
                            if not self.conf_kw["filter_func"](v):

                                #Delete across all variables
                                for cluster_id in fdct[v.as_posix()]:
                                    cluster_var = clusters[cluster_id]
                                    cluster_var.values = [p for p in cluster_var.values if p.as_posix() != v.as_posix()]
                                
                                #Delete the dict entry
                                fdct.pop(v.as_posix(), None)
                            
                            #Otherwise, it's just a normal file, treat it as such
                            else:
                                insert(fdct, v.as_posix(), instance.id)
                                nvals.append(v)

                    elif v not in nvals: 

                        #Add the id 
                        insert(fdct, v.as_posix(), instance.id)
                        nvals.append(v)

                #Add any expected filepath changes

                #print('returned nvals', nvals)
                instance.values = [v for v in list(dict.fromkeys(nvals)) if v.exists() ]
                return func(instance, *args, **kwargs)
            
        


        return wrapper
        
    def set_mode(new_mode):

        if new_mode.lower().strip() not in AutoRefresher.VALID_MODES:
            raise Exception('Invalid mode specified')

        AutoRefresher.REFRESHER_MODE = new_mode
        print('New mode set to ', AutoRefresher.REFRESHER_MODE)
    
    def preserve(self, input_arr):
        '''
        A simple refresher method. This will handle variable conflicts.

        Whenever data is altered, all the invalid paths will drop out.

        Particularly useful for variables with overlapping data
        '''
        return [ v for v in set(input_arr) if v.exists() ]
    
    def get_latest(self, p):

        if AutoRefresher.files_dict[p][1] == None or self.bad_next(p):
            return p 
        
        d = self.get_latest(AutoRefresher.files_dict[p][1])
        if d != p:
            AutoRefresher.files_dict[p][1] = d
        else:
            AutoRefresher.files_dict[p][1] = None

        return d
        
    def bad_next(self, p):
        return AutoRefresher.files_dict[p][1] not in AutoRefresher.files_dict or not AutoRefresher.files_dict[AutoRefresher.files_dict[p][1]][2]

    
    def resolve_file_conflicts(self, input_arr, mapper_func):
        return [self.handle_file_conflict(fl, mapper_func) for fl in set(input_arr)]

    def handle_file_conflict(self, fl, mapper_func):
        '''
            Function to update filnames to resolve 
            potential file conflicts
        '''
        if not mapper_func(fl).exists() or (mapper_func(fl).exists() and mapper_func(fl) == fl):
            return fl
        
        i = 0
        potential_candidate = fl
        while mapper_func(potential_candidate).exists():
            i += 1
            potential_candidate = self.conflict_filename(fl.parent, fl.stem, fl.suffix, i)
            
        fl.rename(potential_candidate)
        return potential_candidate
        
    
    def conflict_filename(self, parent, fl_name, fl_stem, i):
        return parent / f"{fl_name}({i}){fl_stem}"





import fnmatch
import re

class PatternGenerator:
    '''
        The pattern generator is designed to parse all pattern based 
        strings and return regex versions of them that can be used to 
        pattern match.

        We created the generator to help differentiate between glob patterns and
        regex. Most of the time, people use glob, but sometimes the power of regex 
        matching does come in handy. 

        A glob string will come in the normal form. A regex string will start with an r before the actual pattern.

        ex: "a?.txt" is a glob pattern. "ra?.txt" is regex.

    '''
    def is_valid_regex(pattern_string):
        if pattern_string.startswith('r'):
            try:
                re.compile(pattern_string[1:])
            except re.error:
                return False
        else:
            try:
                re.compile(fnmatch.translate(pattern_string))
            except re.error:
                return False

        return True

    def __new__(self, pattern_string):
        '''
            Takes in an input pattern string and returns the appropriate regex pattern.

        '''
        if pattern_string.startswith('r'):
            print('regex')
            return pattern_string[1:]

        #Clear out any leading / trailing spaces 
        print('glob')
        return '' if pattern_string == '' else fnmatch.translate(pattern_string.strip())